import json
import logging

import requests

from . import settings
from .exceptions import CASNetworkError, CASTimeout


logger = logging.getLogger('django')


class AuthApiClient:

    BASE_URL = settings.CAS_BASE_URL
    TIMEOUT_IN_SECONDS = 6

    def __init__(self, cas_credentials=None):
        if cas_credentials is None:
            self._cas_token = settings.CAS_TOKEN
            self._cas_source_id = settings.CAS_SOURCE_ID
        else:
            self._cas_token = cas_credentials['cas_token']
            self._cas_source_id = cas_credentials['cas_source_id']

    def call(self, endpoint, method='GET', payload=None):
        auth_headers = {
            'AUTHORIZATION': 'Token {0}'.format(self._cas_token),
            'SOURCE-ID': self._cas_source_id,
        }
        url = '{base_url}/{endpoint}/.json'.format(
            base_url=self.BASE_URL,
            endpoint=endpoint
        )

        try:
            response = requests.request(
                method,
                url,
                headers=auth_headers,
                json=payload,
                timeout=self.TIMEOUT_IN_SECONDS
            )
        except requests.exceptions.ConnectionError as e:
            raise CASNetworkError from e
        except requests.exceptions.Timeout as e:
            raise CASTimeout from e

        logger.debug('method={0}'.format(method))
        logger.debug('url={0}'.format(url))
        logger.debug('headers={0}'.format(auth_headers))
        logger.debug('payload={0}'.format(payload))
        logger.debug('json={0}'.format(json.dumps(payload)))

        json_data = {}
        try:
            json_data = response.json()
        except ValueError:
            # Requests chokes on empty body
            pass

        return response.status_code, json_data



class CASError(Exception):
    pass


class CASNetworkError(CASError):
    pass


class CASTimeout(CASError):
    pass


class CASUnexpectedStatusCode(CASError):

    def __init__(self, status_code, json):
        message = 'Status: {0}.\nJson: {1}'.format(status_code, json)
        self.status_code = status_code
        self.json = json
        super().__init__(message)


class EmailNotConfirmedError(CASError):
    pass




import json

import responses


def mock_out_post_users(id, email, status=201, **kwargs):
    url = 'https://auth.kagiso.io/api/v1/users/.json'
    data = {
        'id': id,
        'email': email,
        'first_name': kwargs.get('first_name', ''),
        'last_name': kwargs.get('last_name', ''),
        'is_staff': kwargs.get('is_staff', False),
        'is_superuser': kwargs.get('is_superuser', False),
        'confirmation_token': '49:1YkTO2:1VuxvGJre66xqQj6rkEXewmVs08',
        'email_confirmed': None,
        'profile': kwargs.get('profile'),
        'created': '2015-04-21T08:18:30.368602Z',
        'modified': '2015-04-21T08:18:30.374410Z'
    }

    responses.add(
        responses.POST,
        url,
        body=json.dumps(data),
        status=status
    )

    return url, data


def mock_out_put_users(id, email, status=200, **kwargs):
    url = 'https://auth.kagiso.io/api/v1/users/{id}/.json'.format(id=id)
    data = {
        'id': 1,
        'email': email,
        'first_name': kwargs.get('first_name', ''),
        'last_name': kwargs.get('last_name', ''),
        'is_staff': kwargs.get('is_staff', False),
        'is_superuser': kwargs.get('is_superuser', False),
        'profile': kwargs.get('profile'),
        'created': '2015-04-21T08:18:30.368602Z',
        'modified': '2015-04-21T08:18:30.374410Z'
    }

    responses.add(
        responses.PUT,
        url,
        body=json.dumps(data),
        status=status
    )

    return url, data


def mock_out_delete_users(id, status=204):
    url = 'https://auth.kagiso.io/api/v1/users/{id}/.json'.format(id=id)

    responses.add(
        responses.DELETE,
        url,
        status=status
    )

    return url


def mock_out_post_confirm_email(status=200):
    url = 'https://auth.kagiso.io/api/v1/confirm_email/.json'  # noqa

    responses.add(
        responses.POST,
        url,
        status=status
    )

    return url


def mock_out_get_reset_password(email, status=200):
    url = 'https://auth.kagiso.io/api/v1/reset_password/{email}/.json'.format(email=email)  # noqa
    data = {
        'reset_password_token': 'random_token',
    }

    responses.add(
        responses.GET,
        url,
        body=json.dumps(data),
        status=status,
    )

    return url, data


def mock_out_post_reset_password(email, status=200):
    url = 'https://auth.kagiso.io/api/v1/reset_password/{email}/.json'.format(email=email)  # noqa

    responses.add(
        responses.POST,
        url,
        status=status
    )

    return url


def mock_out_post_sessions(status=200, **kwargs):
    url = 'https://auth.kagiso.io/api/v1/sessions/.json'
    data = {
        'id': kwargs.get('id', 1),
        'email': kwargs.get('email'),
        'first_name': kwargs.get('first_name', ''),
        'last_name': kwargs.get('last_name', ''),
        'is_staff': kwargs.get('is_staff', False),
        'is_superuser': kwargs.get('is_superuser', False),
        'email_confirmed': None,
        'profile': kwargs.get('profile'),
        'created': '2015-04-21T08:18:30.368602Z',
        'modified': '2015-04-21T08:18:30.374410Z'
    }

    responses.add(
        responses.POST,
        url,
        body=json.dumps(data),
        status=status
    )

    return url, data


def mock_out_delete_sessions(id, status=200):
    url = 'https://auth.kagiso.io/api/v1/sessions/{id}/.json'.format(id=id)

    responses.add(
        responses.DELETE,
        url,
        status=status
    )

    return url



import os

from django.conf import settings

CAS_TOKEN = os.getenv('CAS_TOKEN') or getattr(
    settings, 'CAS_TOKEN', 'CHANGEME')
CAS_SOURCE_ID = os.getenv('CAS_SOURCE_ID') or getattr(
    settings, 'CAS_SOURCE_ID', 'CHANGEME')
CAS_BASE_URL = os.getenv('CAS_BASE_URL') or getattr(
    settings, 'CAS_BASE_URL', 'https://auth.kagiso.io/api/v1')



from dateutil import parser
from django.db.utils import IntegrityError
from django.test import TestCase
from model_mommy import mommy
import pytest
import responses

from . import mocks
from ... import models
from ...exceptions import CASUnexpectedStatusCode


class KagisoUserTest(TestCase):

    @responses.activate
    def test_create(self):
        # ------------------------
        # -------Arrange----------
        # ------------------------

        email = 'test@email.com'
        first_name = 'Fred'
        last_name = 'Smith'
        is_staff = True
        is_superuser = True
        profile = {
            'age': 22
        }

        url, api_data = mocks.mock_out_post_users(
            1,
            email,
            first_name=first_name,
            last_name=last_name,
            is_staff=is_staff,
            is_superuser=is_superuser,
            profile=profile
        )
        # ------------------------
        # -------Act--------------
        # ------------------------

        user = mommy.make(
            models.KagisoUser,
            id=None,
            email=email,
            first_name=first_name,
            last_name=last_name,
            is_staff=is_staff,
            is_superuser=is_superuser,
            profile=profile,
        )

        # ------------------------
        # -------Assert----------
        # ------------------------

        # Confirmation tokens are saved in memory only.
        assert user.confirmation_token == api_data['confirmation_token']

        result = models.KagisoUser.objects.get(id=user.id)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.url == url

        assert result.id == api_data['id']
        assert result.email == api_data['email']
        assert result.first_name == api_data['first_name']
        assert result.last_name == api_data['last_name']
        assert result.is_staff == api_data['is_staff']
        assert result.is_superuser == api_data['is_superuser']
        assert not result.email_confirmed
        assert result.confirmation_token is None
        assert result.profile == api_data['profile']
        assert result.date_joined == parser.parse(api_data['created'])
        assert result.modified == parser.parse(api_data['modified'])

    @responses.activate
    def test_create_raises_if_user_exists_on_cas(self):
        email = 'test@email.com'
        data = {
            'first_name': 'Fred',
            'last_name': 'Smith',
            'is_staff': True,
            'is_superuser': True,
            'profile': {
                'age': 22,
            }
        }

        url, api_data = mocks.mock_out_post_users(
            1,
            email,
            status=409,
            **data
        )

        with pytest.raises(IntegrityError):
            mommy.make(
                models.KagisoUser,
                id=None,
                email=email,
            )

    @responses.activate
    def test_create_invalid_status_code_raises(self):
        email = 'test@email.com'
        mocks.mock_out_post_users(1, email, status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            mommy.make(
                models.KagisoUser,
                id=None,
                email=email,
            )

    @responses.activate
    def test_update(self):
        # ------------------------
        # -------Arrange----------
        # ------------------------
        mocks.mock_out_post_users(1, 'test@email.com')

        user = mommy.make(models.KagisoUser, id=None)

        email = 'test@email.com'
        first_name = 'Fred'
        last_name = 'Smith'
        is_staff = True
        is_superuser = True
        profile = {
            'age': 22
        }

        url, api_data = mocks.mock_out_put_users(
            1,
            email,
            first_name=first_name,
            last_name=last_name,
            is_staff=is_staff,
            is_superuser=is_superuser,
            profile=profile
        )

        # ------------------------
        # -------Act--------------
        # ------------------------

        user.email = email
        user.first_name = first_name
        user.last_name = last_name
        user.is_staff = is_staff
        user.is_superuser = is_superuser
        user.profile = profile
        user.save()

        # ------------------------
        # -------Assert----------
        # ------------------------
        result = models.KagisoUser.objects.get(id=user.id)

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert result.id == api_data['id']
        assert result.email == api_data['email']
        assert result.first_name == api_data['first_name']
        assert result.last_name == api_data['last_name']
        assert result.is_staff == api_data['is_staff']
        assert result.is_superuser == api_data['is_superuser']
        assert result.profile == api_data['profile']
        assert result.modified == parser.parse(api_data['modified'])

    @responses.activate
    def test_update_syncs_if_user_doesnt_exist_on_cas(self):
        # ------------------------
        # -------Arrange----------
        # ------------------------
        post_url, _ = mocks.mock_out_post_users(1, 'test@email.com')

        user = mommy.make(models.KagisoUser, id=None)

        email = 'test@email.com'
        data = {
            'first_name': 'Fred',
            'last_name': 'Smith',
            'is_staff': True,
            'is_superuser': True,
            'profile': {
                'age': 22,
            }
        }

        put_url, api_data = mocks.mock_out_put_users(
            1,
            email,
            status=404,
        )
        post_url, _ = mocks.mock_out_post_users(2, 'test@email.com', **data)

        # ------------------------
        # -------Act--------------
        # ------------------------

        user.email = email
        user.first_name = data['first_name']
        user.last_name = data['last_name']
        user.is_staff = data['is_staff']
        user.is_superuser = data['is_superuser']
        user.profile = data['profile']
        user.save()

        # ------------------------
        # -------Assert----------
        # ------------------------
        result = models.KagisoUser.objects.get(id=user.id)

        assert len(responses.calls) == 3
        last_api_call = responses.calls[-1].request
        assert last_api_call.method == 'POST'

        assert result.id == api_data['id']
        assert result.email == api_data['email']
        assert result.first_name == api_data['first_name']
        assert result.last_name == api_data['last_name']
        assert result.is_staff == api_data['is_staff']
        assert result.is_superuser == api_data['is_superuser']
        assert result.profile == api_data['profile']
        assert result.modified == parser.parse(api_data['modified'])

    @responses.activate
    def test_update_invalid_status_code_raises(self):
        mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        email = 'test@email.com'
        url, api_data = mocks.mock_out_put_users(1, email, status=500)

        user.email = email

        with pytest.raises(CASUnexpectedStatusCode):
            user.save()

    @responses.activate
    def test_delete(self):
        mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        url = mocks.mock_out_delete_users(user.id)

        user.delete()

        user_deleted = not models.KagisoUser.objects.filter(
            id=user.id).exists()

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert user_deleted

    @responses.activate
    def test_delete_invalid_status_code_raises(self):
        mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        mocks.mock_out_delete_users(user.id, status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            user.delete()

    def test_get_full_name_returns_email(self):
        email = 'test@email.com'
        user = models.KagisoUser(email=email)

        assert user.get_full_name() == email

    def test_get_short_name_returns_email(self):
        email = 'test@email.com'
        user = models.KagisoUser(email=email)

        assert user.get_short_name() == email

    def test_set_password(self):
        user = models.KagisoUser()
        password = 'my_password'

        user.set_password(password)

        assert user.raw_password == password

    def test_get_username_returns_email(self):
        email = 'test@email.com'
        user = models.KagisoUser(email=email)

        assert user.username == email

    def test_set_username_sets_username(self):
        username = 'test@username.com'
        user = models.KagisoUser(username=username)

        assert user.email == username

    @responses.activate
    def test_confirm_email(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        mocks.mock_out_put_users(
            user.id,
            user.email,
            profile=user.profile
        )
        url = mocks.mock_out_post_confirm_email()

        user.confirm_email(post_data['confirmation_token'])

        assert len(responses.calls) == 3
        # Create user, confirm user, update user...
        assert responses.calls[1].request.url == url

        result = models.KagisoUser.objects.get(id=user.id)

        assert result.email_confirmed
        assert not result.confirmation_token

    @responses.activate
    def test_confirm_email_invalid_status_code_raises(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        mocks.mock_out_put_users(
            user.id,
            user.email,
            profile=user.profile
        )
        mocks.mock_out_post_confirm_email(status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            user.confirm_email(post_data['confirmation_token'])

        assert not user.email_confirmed

    @responses.activate
    def test_record_sign_out(self):
        id = 1
        _, post_data = mocks.mock_out_post_users(id, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        url = mocks.mock_out_delete_sessions(id)

        did_sign_out = user.record_sign_out()

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert did_sign_out

    @responses.activate
    def test_record_sign_out_invalid_status_code_raises(self):
        id = 1
        _, post_data = mocks.mock_out_post_users(id, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        mocks.mock_out_delete_sessions(id, status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            did_sign_out = user.record_sign_out()
            assert not did_sign_out

    @responses.activate
    def test_generate_reset_password_token(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        url, data = mocks.mock_out_get_reset_password(user.email)

        reset_password_token = user.generate_reset_password_token()

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert reset_password_token == data['reset_password_token']  # noqa

    @responses.activate
    def test_generate_reset_password_token_invalid_status_raises(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        url, data = mocks.mock_out_get_reset_password(user.email, status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            reset_password_token = user.generate_reset_password_token()
            assert reset_password_token is None

    @responses.activate
    def test_reset_password(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        url = mocks.mock_out_post_reset_password(user.email)

        did_password_reset = user.reset_password('new_password', 'test_token')

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert did_password_reset

    @responses.activate
    def test_reset_password_invalid_status_code_raises(self):
        _, post_data = mocks.mock_out_post_users(1, 'test@email.com')
        user = mommy.make(models.KagisoUser, id=None)
        mocks.mock_out_post_reset_password(user.email, status=500)

        with pytest.raises(CASUnexpectedStatusCode):
            did_password_reset = user.reset_password(
                'new_password',
                'test_token'
            )

            assert not did_password_reset



from unittest.mock import patch

from django.test import TestCase
import pytest
import requests

from ...auth_api_client import AuthApiClient
from ...exceptions import CASNetworkError, CASTimeout


class TestApiClient(TestCase):

    @patch('auth_backend.auth_api_client.requests.request', autospec=True)
    def test_call_raises_on_http_error(self, mock_request):
        auth_api_client = AuthApiClient()
        mock_request.side_effect = requests.exceptions.ConnectionError

        with pytest.raises(CASNetworkError):
            auth_api_client.call('/endpoint/')

    @patch('auth_backend.auth_api_client.requests.request', autospec=True)
    def test_call_raises_on_timeout(self, mock_request):
        auth_api_client = AuthApiClient()
        mock_request.side_effect = requests.exceptions.Timeout

        with pytest.raises(CASTimeout):
            auth_api_client.call('/endpoint/')



from dateutil import parser
from django.contrib.auth.models import AbstractBaseUser, PermissionsMixin
from django.db import models
from django.db.models.signals import pre_delete, pre_save
from django.db.utils import IntegrityError
from django.dispatch import receiver
from django.utils import timezone
from jsonfield import JSONField

from .auth_api_client import AuthApiClient
from .exceptions import CASUnexpectedStatusCode
from .managers import AuthManager


class KagisoUser(AbstractBaseUser, PermissionsMixin):
    USERNAME_FIELD = 'email'

    id = models.IntegerField(primary_key=True)
    email = models.EmailField(max_length=250, unique=True)
    first_name = models.CharField(blank=True, null=True, max_length=100)
    last_name = models.CharField(blank=True, null=True, max_length=100)
    is_staff = models.BooleanField(default=False)
    email_confirmed = models.DateTimeField(null=True)
    profile = JSONField(null=True)
    is_active = models.BooleanField(default=True)
    date_joined = models.DateTimeField()
    modified = models.DateTimeField()

    confirmation_token = None
    raw_password = None

    objects = AuthManager()

    def __init__(self, *args, **kwargs):
        self._auth_api_client = AuthApiClient()
        super().__init__(*args, **kwargs)

    def override_cas_credentials(self, cas_credentials):
        self.cas_credentials = cas_credentials
        self._auth_api_client = AuthApiClient(self.cas_credentials)

    def get_full_name(self):
        return self.email

    def get_short_name(self):
        return self.email

    @property
    def username(self):
        return self.email

    @username.setter
    def username(self, value):
        self.email = value

    def set_password(self, raw_password):
        # We don't want to save passwords locally
        self.set_unusable_password()
        # Save them in memory only
        self.raw_password = raw_password

    def confirm_email(self, confirmation_token):
        payload = {'confirmation_token': confirmation_token}
        endpoint = 'confirm_email'
        status, data = self._auth_api_client.call(endpoint, 'POST', payload)

        if not status == 200:
            raise CASUnexpectedStatusCode(status, data)

        self.confirmation_token = None
        self.email_confirmed = timezone.now()
        self.save()

    def generate_reset_password_token(self):
        endpoint = 'reset_password/{email}'.format(email=self.email)
        status, data = self._auth_api_client.call(endpoint, 'GET')

        if not status == 200:
            raise CASUnexpectedStatusCode(status, data)

        return data['reset_password_token']

    def reset_password(self, password, reset_password_token):
        payload = {
            'reset_password_token': reset_password_token,
            'password': password,
        }
        endpoint = 'reset_password/{email}'.format(email=self.email)
        status, data = self._auth_api_client.call(endpoint, 'POST', payload)

        if not status == 200:
            raise CASUnexpectedStatusCode(status, data)

        return True

    def record_sign_out(self):
        endpoint = 'sessions/{id}'.format(id=self.id)
        status, data = self._auth_api_client.call(endpoint, 'DELETE')

        if not status == 200:
            raise CASUnexpectedStatusCode(status, data)

        return True

    def build_from_cas_data(self, data):
        self.id = data['id']
        self.email = data['email']
        self.first_name = data.get('first_name', self.first_name)
        self.last_name = data.get('last_name', self.last_name)
        self.is_staff = data.get('is_staff', self.is_staff)
        self.is_superuser = data.get('is_superuser', self.is_superuser)
        self.profile = data.get('profile', self.profile)
        self.confirmation_token = data.get('confirmation_token')
        self.date_joined = parser.parse(data['created'])
        self.modified = parser.parse(data['modified'])

    def _create_user_in_db_and_cas(self):
        payload = {
            'email': self.email,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'is_staff': self.is_staff,
            'is_superuser': self.is_superuser,
            'profile': self.profile,
            'password': self.raw_password,
        }

        status, data = self._auth_api_client.call('users', 'POST', payload)

        if status not in (201, 409):
            raise CASUnexpectedStatusCode(status, data)

        # 409-Conflict means that the user already exists in CAS
        if status == 409:
            raise IntegrityError('User already exists')

        self.build_from_cas_data(data)

    def _update_user_in_cas(self):
        payload = {
            'email': self.email,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'is_staff': self.is_staff,
            'is_superuser': self.is_superuser,
            'profile': self.profile,
        }

        status, data = self._auth_api_client.call(
            'users/{id}'.format(id=self.id), 'PUT', payload)

        if status == 200:
            self.email = data['email']
            self.first_name = data.get('first_name')
            self.last_name = data.get('last_name')
            self.is_staff = data.get('is_staff')
            self.is_superuser = data.get('is_superuser')
            self.profile = data.get('profile')
            self.modified = parser.parse(data['modified'])
        elif status == 404:
            # It is possible that a user exists locally but not on CAS
            # eg. when converting an existing app to use CAS
            # so on update if the user is not found, then create a CAS user
            self._create_user_in_db_and_cas()
        else:
            raise CASUnexpectedStatusCode(status, data)

    def __str__(self):
        return self.email  # pragma: no cover


@receiver(pre_delete, sender=KagisoUser)
def delete_user_from_cas(sender, instance, *args, **kwargs):
    status, data = instance._auth_api_client.call(
        'users/{id}'.format(id=instance.id), 'DELETE')

    # It is possible but unlikely that a user exists locally but not on CAS
    # eg. when converting an existing app to use CAS
    # So if not found on CAS just proceed to delete locally
    if status not in (204, 404):
        raise CASUnexpectedStatusCode(status, data)


@receiver(pre_save, sender=KagisoUser)
def save_user_to_cas(sender, instance, *args, **kwargs):
    if not instance.id:
        instance._create_user_in_db_and_cas()
    else:
        instance._update_user_in_cas()



from django.test import TestCase
from model_mommy import mommy
import pytest

from . import utils
from ... import models
from ...backends import KagisoBackend
from ...exceptions import EmailNotConfirmedError


class KagisoUserTest(TestCase):

    def test_user(self):
        # ----- Create user -----
        email = utils.random_email()
        profile = {
            'age': 15,
        }
        password = 'my_password'

        user = mommy.prepare(
            models.KagisoUser,
            id=None,
            email=email,
            first_name='George',
            last_name='Smith',
            is_staff=True,
            is_superuser=True,
            profile=profile,
        )
        user.set_password(password)
        user.save()

        auth_backend = KagisoBackend()

        result = models.KagisoUser.objects.get(id=user.id)

        assert result.email == user.email
        assert result.first_name == user.first_name
        assert result.last_name == user.last_name
        assert result.is_staff
        assert result.is_superuser
        assert not result.confirmation_token
        assert result.profile == profile

        # ----- Unconfirmed users can't sign in -----
        with pytest.raises(EmailNotConfirmedError):
            auth_backend.authenticate(
                email=user.email,
                password=password
            )

        # ----- Confirm user -----
        assert not user.email_confirmed
        user.confirm_email(user.confirmation_token)
        assert user.email_confirmed

        # ----- Can the user sign in? -----
        signed_in_user = auth_backend.authenticate(
            email=user.email,
            password=password
        )
        assert signed_in_user == user

        # ----- Update user -----
        new_email = utils.random_email()
        new_first_name = 'Fred'
        new_last_name = 'Jones'
        new_profile = {
            'age': 50,
        }

        user.email = new_email
        user.first_name = new_first_name
        user.last_name = new_last_name
        user.is_staff = False
        user.is_superuser = False
        user.profile = new_profile
        user.save()

        updated_result = models.KagisoUser.objects.get(id=user.id)

        assert updated_result.email == new_email
        assert updated_result.first_name == new_first_name
        assert updated_result.last_name == new_last_name
        assert not updated_result.is_staff
        assert not updated_result.is_superuser
        assert updated_result.profile == new_profile

        # ----- Sign out the user -----
        signed_out = user.record_sign_out()
        assert signed_out

        # ----- Delete user -----
        updated_result.delete()

    def test_social_sign_in(self):
        # ----- Create user -----
        email = utils.random_email()

        user = mommy.prepare(
            models.KagisoUser,
            id=None,
            email=email,
        )
        user.set_unusable_password()
        user.save()
        user.confirm_email(user.confirmation_token)

        # ----- Can the user sign in? -----
        auth_backend = KagisoBackend()
        signed_in_user = auth_backend.authenticate(
            email=user.email,
            strategy='facebook'
        )

        assert signed_in_user == user

    def test_reset_password(self):
        # ----- Create the user -----
        email = utils.random_email()

        user = mommy.prepare(
            models.KagisoUser,
            id=None,
            email=email,
        )
        user.set_password('password')
        user.save()
        user.confirm_email(user.confirmation_token)

        # ----- Reset the password -----
        new_password = 'new_password'
        token = user.generate_reset_password_token()
        did_password_reset = user.reset_password(new_password, token)
        assert did_password_reset

        # ----- Can the user still sign in? -----
        auth_backend = KagisoBackend()
        result = auth_backend.authenticate(
            email=user.email,
            password=new_password
        )
        assert result == user



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils.timezone import utc
import datetime


class Migration(migrations.Migration):

    dependencies = [
        ('auth_backend', '0003_kagisouser_profile'),
    ]

    operations = [
        migrations.AddField(
            model_name='kagisouser',
            name='modified',
            field=models.DateTimeField(default=datetime.datetime(2015, 4, 22, 9, 3, 27, 81025, tzinfo=utc)),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='kagisouser',
            name='date_joined',
            field=models.DateTimeField(),
            preserve_default=True,
        ),
    ]



from django.contrib.auth.backends import ModelBackend
from django.db.models.signals import pre_save

from .auth_api_client import AuthApiClient

from .exceptions import CASUnexpectedStatusCode, EmailNotConfirmedError
from .models import KagisoUser, save_user_to_cas


class KagisoBackend(ModelBackend):

    # Django calls our backend with username='xyz', password='abc'
    # e.g. credentials = {'username': 'Fred', 'password': 'open'}
    # authenticate(**credentials), even though we set USERNAME_FIELD to
    # 'email' in models.py.
    #
    # Django AllAuth does this:
    #  credentials = {'email': 'test@kagiso.io, 'password': 'open'}
    def authenticate(self, email=None, username=None, password=None, **kwargs):
        cas_credentials = kwargs.get('cas_credentials')

        email = username if not email else email
        existing_user = KagisoUser.objects.filter(email=email).first()

        payload = {
            'email': email,
        }

        # Social signins don't have passwords
        if password:
            payload['password'] = password

        # Support social sign_ins
        strategy = kwargs.get('strategy')
        if strategy:
            payload['strategy'] = strategy

        auth_api_client = AuthApiClient(cas_credentials)
        status, data = auth_api_client.call('sessions', 'POST', payload)

        if status == 200:
            if existing_user:
                existing_user.override_cas_credentials(cas_credentials)
            else:
                try:
                    # Do not on save sync to CAS, as we just got the user's
                    # data from CAS, and nothing has changed in the interim
                    pre_save.disconnect(save_user_to_cas, sender=KagisoUser)
                    existing_user = KagisoUser()
                    existing_user.set_password(password)
                    existing_user.build_from_cas_data(data)
                    existing_user.save()
                finally:
                    pre_save.connect(save_user_to_cas, sender=KagisoUser)
        elif status == 404:
            return None
        elif status == 422:
            raise EmailNotConfirmedError()
        else:
            raise CASUnexpectedStatusCode(status, data)

        return existing_user



from django.test import TestCase
import pytest
import responses

from . import mocks
from ...backends import KagisoBackend
from ...exceptions import CASUnexpectedStatusCode, EmailNotConfirmedError
from ...models import KagisoUser


class KagisoBackendTest(TestCase):

    @responses.activate
    def test_authenticate_valid_credentials_returns_user(self):
        email = 'test@email.com'
        password = 'random'
        profile = {
            'first_name': 'Fred'
        }
        mocks.mock_out_post_users(
            1,
            email,
            profile=profile
        )
        user = KagisoUser.objects.create_user(
            email, password, profile=profile)
        url, _ = mocks.mock_out_post_sessions(200)

        backend = KagisoBackend()
        result = backend.authenticate(email=email, password=password)

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert isinstance(result, KagisoUser)
        assert result.id == user.id

    @responses.activate
    def test_authenticate_valid_credentials_creates_local_user_if_none(self):
        email = 'test@email.com'
        password = 'random'

        data = {
            'id': 55,
            'email': email,
            'first_name': 'Fred',
            'last_name': 'Smith',
            'is_staff': True,
            'is_superuser': True,
            'profile': {'age': 40, },
        }

        _, api_data = mocks.mock_out_post_users(1, email)
        session_url, data = mocks.mock_out_post_sessions(200, **data)

        backend = KagisoBackend()
        result = backend.authenticate(email=email, password=password)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.url == session_url

        assert result.id == data['id']
        assert result.email == data['email']
        assert result.first_name == data['first_name']
        assert result.last_name == data['last_name']
        assert result.is_staff == data['is_staff']
        assert result.is_superuser == data['is_superuser']
        assert result.profile == data['profile']

    @responses.activate
    def test_authenticate_invalid_status_code_raises(self):
        email = 'test@email.com'
        password = 'random'
        url, api_data = mocks.mock_out_post_users(1, email)
        KagisoUser.objects.create_user(email, password)

        mocks.mock_out_post_sessions(500)

        backend = KagisoBackend()

        with pytest.raises(CASUnexpectedStatusCode):
            backend.authenticate(email=email, password=password)

    @responses.activate
    def test_authenticate_with_social_sign_in_returns_user(self):
        email = 'test@email.com'
        strategy = 'facebook'
        mocks.mock_out_post_users(1, email)
        # Unusable password is saved locally for Django compliance
        # It is not used for auth purposes though
        user = KagisoUser.objects.create_user(email, password='unusable')
        url, data = mocks.mock_out_post_sessions(200)

        backend = KagisoBackend()
        result = backend.authenticate(email=email, strategy=strategy)

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert isinstance(result, KagisoUser)
        assert result.id == user.id

    @responses.activate
    def test_authenticate_invalid_credentials_returns_none(self):
        email = 'test@email.com'
        password = 'incorrect'
        mocks.mock_out_post_users(1, email)
        KagisoUser.objects.create_user(email, password)
        url, data = mocks.mock_out_post_sessions(404)

        backend = KagisoBackend()
        result = backend.authenticate(email=email, password=password)

        assert len(responses.calls) == 2
        assert responses.calls[1].request.url == url

        assert not result

    @responses.activate
    def test_authenticate_unconfirmed_email_raises(self):
        email = 'test@email.com'
        password = 'random'
        url, api_data = mocks.mock_out_post_users(1, email)
        KagisoUser.objects.create_user(email, password)

        mocks.mock_out_post_sessions(422)

        backend = KagisoBackend()

        with pytest.raises(EmailNotConfirmedError):
            backend.authenticate(email=email, password=password)



import os

SECRET_KEY = '^7jcx7^h#b@%a76lr@a2!7xj#@4@5ayuyan9c$y#(_(8l3)_%t'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'auth_backend',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'auth_backend.urls'

WSGI_APPLICATION = 'auth_backend.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'auth_backend',
        'USER': os.getenv('USER'),
        'PASSWORD': 'password',
        'ATOMIC_REQUESTS': True,
    }
}


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
    },
}


# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_URL = '/static/'

AUTH_USER_MODEL = 'auth_backend.KagisoUser'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('auth_backend', '0004_auto_20150422_0903'),
    ]

    operations = [
        migrations.AddField(
            model_name='kagisouser',
            name='email_confirmed',
            field=models.DateTimeField(null=True),
            preserve_default=True,
        ),
    ]



from dateutil import parser
from django.test import TestCase
import responses

from . import mocks
from ...models import KagisoUser


class KagisoUserTest(TestCase):

    @responses.activate
    def test_create_user(self):
        email = 'test@email.com'
        first_name = 'Fred'
        last_name = 'Smith'
        is_staff = True
        is_superuser = True
        password = 'random'
        profile = {
            'age': 18,
        }
        url, api_data = mocks.mock_out_post_users(
            1,
            email,
            first_name=first_name,
            last_name=last_name,
            is_staff=is_staff,
            is_superuser=is_superuser,
            profile=profile
        )

        result = KagisoUser.objects.create_user(
            email, password, profile=profile)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.url == url

        assert result.id == api_data['id']
        assert result.email == email
        assert result.first_name == first_name
        assert result.last_name == last_name
        assert result.is_staff == is_staff
        assert result.is_superuser == is_superuser
        assert result.confirmation_token == api_data['confirmation_token']
        assert not result.email_confirmed
        assert result.profile == profile
        assert result.date_joined == parser.parse(api_data['created'])
        assert result.modified == parser.parse(api_data['modified'])

    @responses.activate
    def test_create_super_user(self):
        email = 'test@email.com'
        password = 'random'
        url, api_data = mocks.mock_out_post_users(1, email, is_superuser=True)

        result = KagisoUser.objects.create_superuser(email, password)

        assert len(responses.calls) == 1
        assert responses.calls[0].request.url == url

        assert result.email == email
        assert result.is_superuser
